# Collection of citations for modules in GPUBACKENDTOOLS package

"""
:code:`gpubackendtools.utils.citations`:

This module is used to collect citations for all modules in the package. This
module is then imported to add citations to module classes using their :code:`citation`
attribute.
"""

import abc
import enum
from typing import Iterable, List, Optional, Union

from pydantic import BaseModel

from gpubackendtools.utils.exceptions import InvalidInputFile


def _hyphen_replace(field: str) -> str:
    return field.replace("_", "-")


class HyphenUnderscoreAliasModel(BaseModel):
    """Pydantic model were hyphen replace underscore in field names."""

    class Config:
        alias_generator = _hyphen_replace
        extra = "ignore"
        frozen = True


class Author(HyphenUnderscoreAliasModel):
    """Description of a reference author."""

    family_names: str
    given_names: str
    orcid: Optional[str] = None
    affiliation: Optional[str] = None
    email: Optional[str] = None


class Publisher(HyphenUnderscoreAliasModel):
    """Description of a publisher."""

    name: str


class Identifier(HyphenUnderscoreAliasModel):
    """Description of an identifier."""

    type: str
    value: str
    description: Optional[str] = None


class ReferenceABC(HyphenUnderscoreAliasModel, abc.ABC):
    """Abstract base class for references."""

    @abc.abstractmethod
    def to_bibtex(self) -> str:
        """Convert a reference object to a BibTeX string representation."""


class ArxivIdentifier(BaseModel):
    """Class representing an arXiv identifier"""

    reference: str
    primary_class: Optional[str] = None


class ArticleReference(ReferenceABC):
    """Description of an article."""

    abbreviation: str
    authors: List[Author]
    title: str
    journal: Optional[str] = None
    year: int
    month: Optional[int] = None
    issue: Optional[int] = None
    publisher: Optional[Publisher] = None
    pages: Optional[int] = None
    start: Optional[int] = None
    issn: Optional[str] = None
    doi: Optional[str] = None
    identifiers: Optional[List[Identifier]] = None

    @property
    def arxiv_preprint(self) -> Optional[ArxivIdentifier]:
        """
        Detect an arXiv identifier if any.

        an arXiv identifier is:
        - an identifier of type "other"
        - which starts with "arxiv:" (case insensitive)
        - whose second part is either:
          - The arXiv reference (e.g. "arxiv:1912.07609")
          - The primary class followed by '/' and the reference (e.g. "arxiv:gr-qc/1912.07609")
        """
        if self.identifiers is None:
            return None

        for identifier in self.identifiers:
            if identifier.type != "other":
                continue
            if not identifier.value.lower().startswith("arxiv:"):
                continue
            data = identifier.value.lower().removeprefix("arxiv:")
            primary_class, reference = (
                data.split("/", 1) if "/" in data else (None, data)
            )
            return ArxivIdentifier(primary_class=primary_class, reference=reference)
        return None

    def to_bibtex(self) -> str:
        """Build the BibTeX representation of an article."""
        arxiv_id = self.arxiv_preprint

        line_format = (
            """  {:<10} = \"{}\"""" if arxiv_id is None else """  {:<13} = \"{}\""""
        )

        def format_line(key: str, value: str, format: str = line_format) -> str:
            return format.format(key, value)

        lines = []
        lines.append("@article{" + self.abbreviation)
        lines.append(
            format_line(
                "author",
                " and ".join(
                    [
                        "{}, {}".format(author.family_names, author.given_names)
                        for author in self.authors
                    ]
                ),
            )
        )
        lines.append(format_line("title", "{" + self.title + "}"))
        if self.journal is not None:
            lines.append(format_line("journal", self.journal))
        lines.append(format_line("year", str(self.year)))
        if self.month is not None:
            lines.append(format_line("month", str(self.month)))
        if self.issue is not None:
            lines.append(format_line("number", str(self.issue)))
        if self.publisher is not None:
            lines.append(format_line("publisher", str(self.publisher.name)))
        if self.start is not None:
            lines.append(
                format_line(
                    "pages",
                    str(self.start)
                    if self.pages is None
                    else "{}--{}".format(self.start, self.start + self.pages),
                )
            )
        if self.issn is not None:
            lines.append(format_line("issn", str(self.issn)))
        if self.doi is not None:
            lines.append(format_line("doi", str(self.doi)))
        if arxiv_id is not None:
            lines.append(format_line("archivePrefix", "arXiv"))
            lines.append(format_line("eprint", arxiv_id.reference))
            if arxiv_id.primary_class is not None:
                lines.append(format_line("primaryClass", arxiv_id.primary_class))

        return ",\n".join(lines) + "\n}"


class SoftwareReference(ReferenceABC):
    """Description of a Software"""

    authors: list[Author]
    title: str

    license: Optional[str] = None
    url: Optional[str] = None
    repository: Optional[str] = None
    identifiers: Optional[List[Identifier]] = None
    year: Optional[int] = None
    month: Optional[int] = None
    version: Optional[str] = None

    @property
    def doi(self) -> Optional[str]:
        """Return the first DOI in identifiers if any"""
        if self.identifiers is None:
            return None

        for identifier in self.identifiers:
            if identifier.type == "doi":
                return identifier.value
        return None

    def to_bibtex(self) -> str:
        """Build the BibTeX representation of a software."""

        def format_line(key: str, value: str) -> str:
            return """  {:<10} = \"{}\"""".format(key, value)

        lines = []
        lines.append("@software{" + self.title)
        lines.append(
            format_line(
                "author",
                " and ".join(
                    [
                        "{}, {}".format(author.family_names, author.given_names)
                        for author in self.authors
                    ]
                ),
            )
        )
        lines.append(format_line("title", "{" + self.title + "}"))
        if self.license is not None:
            lines.append(format_line("license", self.license))
        if self.url is not None:
            lines.append(format_line("url", self.url))
        if self.repository is not None:
            lines.append(format_line("repository", self.repository))
        if self.year is not None:
            lines.append(format_line("year", str(self.year)))
        if self.month is not None:
            lines.append(format_line("month", str(self.month)))
        if self.version is not None:
            lines.append(format_line("version", self.version))
        if self.doi is not None:
            lines.append(format_line("doi", str(self.doi)))

        return ",\n".join(lines) + "\n}"


Reference = Union[ArticleReference, SoftwareReference]


class REFERENCE(enum.Enum):
    FAST_GB = "Cornish:2007if"
    TRIPLES_1 = "Robson:2018svj"
    GPUBACKENDTOOLS_SOFTWARE = "michael_l_katz_2022_6500434"

    def __str__(self) -> str:
        return str(self.value)


class CitationRegistry:
    __slots__ = "registry"

    registry: dict[str, Reference]

    def __init__(self, **kwargs):
        self.registry = kwargs

    def get(self, key: Union[str, REFERENCE]) -> Reference:
        """Return a Reference object from its key."""
        return self.registry[key if isinstance(key, str) else key.value]

    def all(self) -> Iterable[Reference]:
        return self.registry.values()


def build_citation_registry() -> CitationRegistry:
    """Read the package CITATION.cff and build the corresponding registry."""
    import json
    import pathlib

    import jsonschema
    import yaml

    from gpubackendtools import __file__ as _gpubackendtools_root_file
    from gpubackendtools import _is_editable as is_editable

    gpubackendtools_root = pathlib.Path(_gpubackendtools_root_file).parent
    cff_root = (
        gpubackendtools_root.parent.parent if is_editable else gpubackendtools_root
    )
    citation_cff_path = cff_root / "CITATION.cff"

    with open(citation_cff_path, "rt") as fid:
        cff = yaml.safe_load(fid)

    with open(pathlib.Path(__file__).parent / "cff_1_2_0.schema.json", "r") as f:
        cff_schema = json.load(f)

    try:
        jsonschema.validate(cff, cff_schema)
    except jsonschema.SchemaError as e:
        raise InvalidInputFile("cff_1_2_0.schema.json is not a valid schema.") from e
    except jsonschema.exceptions.ValidationError as e:
        raise InvalidInputFile(
            "The file {} does not match its expected schema. Contact gpubackendtools developers.".format(
                citation_cff_path
            )
        ) from e

    def to_reference(ref_dict) -> Reference:
        if ref_dict["type"] == "article":
            return ArticleReference(**ref_dict)
        if ref_dict["type"] == "software":
            return SoftwareReference(**ref_dict)

        raise InvalidInputFile(
            "The file {} contains references whose type ({}) ".format(
                citation_cff_path, ref_dict["type"]
            )
            + "is not supported",
            ref_dict,
        )

    references = {ref["abbreviation"]: to_reference(ref) for ref in cff["references"]}

    return CitationRegistry(**references, **{cff["title"]: to_reference(cff)})


COMMON_REFERENCES = [
    REFERENCE.FAST_GB,
    REFERENCE.TRIPLES_1,
    REFERENCE.GPUBACKENDTOOLS_SOFTWARE,
]


class Citable:
    """Base class for classes associated with specific citations."""

    registry: Optional[CitationRegistry] = None

    @classmethod
    def citation(cls) -> str:
        """Return the module references as a printable BibTeX string."""
        references = cls.module_references()

        registry = cls._get_registry()

        bibtex_entries = [registry.get(str(key)).to_bibtex() for key in references]
        return "\n\n".join(bibtex_entries)

    @classmethod
    def all_citations(cls) -> str:
        """Return all the citations from the registry as printable BibTeX string"""
        registry = cls._get_registry()

        bibtex_entries = [entry.to_bibtex() for entry in registry.all()]
        return "\n\n".join(bibtex_entries)

    @classmethod
    def module_references(cls) -> Iterable[Union[REFERENCE, str]]:
        """Method implemented by each class to define its list of references"""
        return COMMON_REFERENCES

    @classmethod
    def _get_registry(cls) -> CitationRegistry:
        if Citable.registry is None:
            from gpubackendtools import get_logger

            get_logger().debug("Building the Citation Registry from CITATION.cff")
            Citable.registry = build_citation_registry()
        return Citable.registry



from .globals import get_backend, get_first_backend
from .gpubackendtools import Backend
from typing import Optional, Sequence, TypeVar
import types


class ParallelModuleBase:  # (Citable):
    """
    Base class for modules that can use a GPU (or revert back to CPU).

    This class mainly handles backend selection. Each backend offers accelerated
    computations on a specific device (cpu, CUDA 11.x enabled GPU, CUDA 12.x enabled GPU).

    args:
        force_backend (str, optional): Name of the backend to use
    """

    _backend_name: str

    def __init__(self, /, force_backend=None):
        if force_backend is None:
            self._backend_name = get_first_backend(self.supported_backends()).name

        elif isinstance(force_backend, str):
            self._backend_name = get_backend(force_backend).name
        elif isinstance(force_backend, Backend):
            force_backend = force_backend.name
            self._backend_name = get_backend(force_backend).name
        elif isinstance(force_backend, tuple) or isinstance(force_backend, list):
            # comes in the form of (module, backend)
            assert isinstance(force_backend[0], str) and isinstance(
                force_backend[1], str
            )
            force_backend = force_backend[0] + "_" + force_backend[1]
            self._backend_name = get_backend(force_backend).name

        else:
            raise ValueError("FIX THIS")

    @property
    def backend(self) -> Backend:
        """Access the underlying backend."""
        return get_backend(self._backend_name)

    @classmethod
    def supported_backends(cls) -> Sequence[str]:
        """List of backends supported by a parallel module by order of preference."""
        raise NotImplementedError(
            "Class {} does not implement the supported_backends method.".format(cls)
        )

    @staticmethod
    def CPU_ONLY() -> list[str]:
        """List of supported backend for CPU only class"""
        return ["cpu"]

    @staticmethod
    def GPU_RECOMMENDED() -> list[str]:
        """List of supported backends for GPU-recommended class with CPU support"""
        return ["cuda12x", "cuda11x", "cpu"]

    @staticmethod
    def CPU_RECOMMENDED_WITH_GPU_SUPPORT() -> list[str]:
        """List of supported backends for CPU-recommended class with GPU support"""
        return ["cpu", "cuda12x", "cuda11x"]

    @staticmethod
    def GPU_ONLY() -> list[str]:
        """List of supported backends for GPU-only class"""
        return ["cuda12x", "cuda11x"]

    @property
    def xp(self) -> types.ModuleType:
        """Return the module providing ndarray capabilities"""
        return self.backend.xp

    @property
    def backend_name(self) -> str:
        """Return the name of current backend"""
        return self.backend.name

    ParallelModuleDerivate = TypeVar(
        "ParallelModuleDerivate", bound="ParallelModuleBase"
    )

    def build_with_same_backend(
        self,
        module_class: type[ParallelModuleDerivate],
        args: Optional[list] = None,
        kwargs: Optional[dict] = None,
    ) -> ParallelModuleDerivate:
        """
        Build an instance of `module_class` with same backend as current object.

        args:
          module_class: class of the object to be built, must derive from ParallelModuleBase
          args (list, optional): positional arguments for module_class constructor
          kwargs (dict, optional): keyword arguments for module_class constructor
                                   (the 'force_backend' argument will be ignored and replaced)
        """
        args = [] if args is None else args
        return module_class(*args, **self.adapt_backend_kwargs(kwargs=kwargs))

    def adapt_backend_kwargs(self, kwargs: Optional[dict] = None) -> dict:
        """Adapt a set of keyword arguments to add/set 'force_backend' to current backend"""
        if kwargs is None:
            kwargs = {}
        kwargs["force_backend"] = self.backend_name
        return kwargs



"""Implementation of a centralized configuration management for GPUBACKENDTOOLS."""

from __future__ import annotations

import abc
import argparse
import dataclasses
import enum
import logging
import os
import pathlib
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from .. import exceptions


class ConfigSource(enum.Enum):
    """Enumeration of config option sources."""

    DEFAULT = "default"
    """Config value comes from its default value"""

    CFGFILE = "config_file"
    """Config value comes from the configuration file"""

    ENVVAR = "environment_var"
    """Config value comes from environment variable"""

    CLIOPT = "command_line"
    """Config value comes from command line parameter"""

    SETTER = "setter"
    """Config value set by config setter after importing GPUBACKENDTOOLS"""


T = TypeVar("T")

ENVVAR_PREFIX: str = "GPUBACKENDTOOLS_"


@dataclasses.dataclass
class ConfigEntry(Generic[T]):
    """Description of a configuration entry."""

    label: str
    """How the entry is referred to in Python code (config.get("label"))"""
    description: str
    """Description of the configuration entry"""

    type: TypeVar = T
    """Type of the value"""

    default: Optional[T] = None
    """Default value"""

    cfg_entry: Optional[str] = None
    """Name of the entry in a config file"""

    env_var: Optional[str] = None
    """Entry corresponding env var (uppercase, without GPUBACKENDTOOLS_ prefix)"""

    cli_flags: Optional[Union[str, List[str]]] = None
    """Flag(s) for CLI arguments of this entry (e.g. "-f")"""

    cli_kwargs: Optional[Dict[str, Any]] = None
    """Supplementary arguments to argparse add_argument method for CLI options"""

    convert: Callable[[str], T] = None
    """Method used to convert a user input to expected type"""

    validate: Callable[[T], bool] = lambda _: True
    """Method used to validate the provided option value"""

    overwrite: Callable[[T, T], T] = lambda _, new: new
    """Method used to update the value if given by multiple means"""

    def __post_init__(self):
        if self.convert is None:
            self.convert = lambda v: self.type(v)


def compatibility_isinstance(obj, cls) -> bool:
    import sys
    import typing

    if (sys.version_info >= (3, 10)) or (typing.get_origin(cls) is None):
        try:
            return isinstance(obj, cls)
        except TypeError:
            pass

    if typing.get_origin(cls) is typing.Union:
        for arg in typing.get_args(cls):
            if compatibility_isinstance(obj, arg):
                return True
        return False

    if typing.get_origin(cls) is list:
        if not isinstance(obj, list):
            return False
        for item in obj:
            if not compatibility_isinstance(item, typing.get_args(cls)[0]):
                return False
        return True

    import collections.abc

    if typing.get_origin(cls) is collections.abc.Sequence:
        if not hasattr(obj, "__iter__"):
            return False
        for item in obj:
            if not compatibility_isinstance(item, typing.get_args(cls)[0]):
                return False
        return True

    raise NotImplementedError(
        "Compatiblity wrapper for isinstance on Python 3.9 does not support given type."
    )


@dataclasses.dataclass
class ConfigItem(Generic[T]):
    """Actual configuration entry with its run-time properties (value, source, ...)"""

    value: T  # Item value
    source: ConfigSource  # Source of the item current value


class ConfigConsumer(abc.ABC):
    """
    Base class for actual configs.

    This class handles building config values from default, file, environment and CLI options.
    It keeps a list of unused parameters for an other consumer.
    """

    _entries: Dict[str, ConfigEntry]
    _items: Dict[str, ConfigItem]

    _extra_file: Dict[str, str]
    _extra_env: Dict[str, str]
    _extra_cli: List[str]

    @classmethod
    @abc.abstractmethod
    def config_entries(cls) -> List[ConfigEntry]:
        """Return the list of the class config entries"""
        raise NotImplementedError(
            "A ConfigConsumer must implement 'config_entries' method."
        )

    def __init__(
        self,
        config_file: Union[os.PathLike, Mapping[str, str], None] = None,
        env_vars: Optional[Mapping[str, str]] = None,
        cli_args: Optional[Sequence[str]] = None,
        set_args: Optional[Dict[str, Any]] = None,
    ):
        """Initialize the items list and extra parameters."""
        config_entries = self.config_entries()

        # Build the entries mapping
        self._entries = {entry.label: entry for entry in config_entries}

        # Build default items
        default_items = ConfigConsumer._build_items_from_default(config_entries)

        # Retrieve option from sources
        opt_from_file = ConfigConsumer._get_from_config_file(config_file)
        opt_from_env = ConfigConsumer._get_from_envvars(env_vars)
        opt_from_cli = ConfigConsumer._get_from_cli_args(cli_args)

        # Consume options to build other item lists
        file_items, self._extra_file = ConfigConsumer._build_items_from_file(
            config_entries, opt_from_file
        )
        env_items, self._extra_env = ConfigConsumer._build_items_from_env(
            config_entries, opt_from_env
        )
        cli_items, self._extra_cli = ConfigConsumer._build_items_from_cli(
            config_entries, opt_from_cli
        )
        set_items = ConfigConsumer._build_items_from_set(config_entries, set_args)

        # Build final item mapping
        self._items = self._overwrite(
            default_items, file_items, env_items, cli_items, set_items
        )

        # Validate items:
        errors: List[Exception] = []
        for label, entry in self._entries.items():
            if label not in self._items:
                errors.append(
                    exceptions.ConfigurationMissing(
                        "Configuration entry '{}' is missing.".format(label)
                    )
                )
                continue
            item = self._items[label]
            if not entry.validate(item.value):
                errors.append(
                    exceptions.ConfigurationValidationError(
                        "Configuration entry '{}' has invalid value '{}'".format(
                            label, item.value
                        )
                    )
                )

        if errors:
            raise (
                errors[0]
                if len(errors) == 1
                else exceptions.ExceptionGroup(
                    "Invalid configuration due to previous issues.", errors
                )
            )

    def _overwrite(
        self,
        old_items: Dict[str, ConfigItem],
        new_items: Dict[str, ConfigItem],
        *other_items,
    ) -> Dict[str, ConfigItem]:
        merged_items = {}
        for label, old_item in old_items.items():
            if label not in new_items:
                merged_items[label] = old_item

        for label, new_item in new_items.items():
            if label not in old_items:
                merged_items[label] = new_item
                continue
            old_item = old_items[label]
            entry = self._entries[label]
            merged_items[label] = ConfigItem(
                value=entry.overwrite(old_item.value, new_item.value),
                source=new_item.source,
            )
        if len(other_items) > 0:
            return self._overwrite(merged_items, *other_items)
        return merged_items

    def __getitem__(self, key: str) -> Any:
        """Get the value of a config entry."""
        return self._items[key].value

    def __getattr__(self, attr: str) -> Any:
        """Get the value of a config entry via attributes."""
        if attr in self._items:
            return self._items[attr].value
        return self.__getattribute__(attr)

    def get_item(self, key: str) -> Tuple[ConfigItem, ConfigEntry]:
        return self._items[key], self._entries[key]

    def get_items(self) -> List[Tuple[ConfigItem, ConfigEntry]]:
        return [(self._items[key], entry) for key, entry in self._entries.items()]

    def get_extras(self) -> Tuple[Mapping[str, str], Mapping[str, str], Sequence[str]]:
        """Return extra file, env and cli entries for other consumer."""
        return self._extra_file, self._extra_env, self._extra_cli

    @staticmethod
    def _get_from_config_file(
        config_file: Union[os.PathLike, Mapping[str, str], None],
    ) -> Dict[str, str]:
        """Read a config file and return its items as a dictionary."""
        if config_file is None:
            return {}
        if isinstance(config_file, Mapping):
            return {key: value for key, value in config_file.items()}
        import configparser

        config = configparser.ConfigParser()
        config.read(config_file)
        if "gpubackendtools" not in config:
            return {}
        gpubackendtools_section = config["gpubackendtools"]
        return {key: gpubackendtools_section[key] for key in gpubackendtools_section}

    @staticmethod
    def _get_from_envvars(env_vars: Optional[Mapping[str, str]]) -> Dict[str, str]:
        """Filter-out environment variables not matching a given prefix."""
        return (
            {
                key: value
                for key, value in env_vars.items()
                if key.startswith(ENVVAR_PREFIX)
            }
            if env_vars is not None
            else {}
        )

    @staticmethod
    def _get_from_cli_args(cli_args: Optional[Sequence[str]]) -> List[str]:
        """Build list of CLI arguments."""
        return [arg for arg in cli_args] if cli_args is not None else []

    @staticmethod
    def _build_items_from_default(
        config_entries: Sequence[ConfigEntry],
    ) -> Dict[str, ConfigItem]:
        """Build a list of ConfigItem built from default-valued ConfigEntries."""
        return {
            entry.label: ConfigItem(value=entry.default, source=ConfigSource.DEFAULT)
            for entry in config_entries
            if compatibility_isinstance(entry.default, entry.type)
        }

    @staticmethod
    def _build_items_from_file(
        config_entries: Sequence[ConfigEntry], opt_from_file: Mapping[str, str]
    ) -> Tuple[Dict[str, ConfigItem], Dict[str, str]]:
        """Extract configuration items from file option and build dict of unconsumed extra items."""
        extras_from_file = {**opt_from_file}
        items_from_file = {
            entry.label: ConfigItem(
                value=entry.convert(extras_from_file.pop(entry.cfg_entry)),
                source=ConfigSource.CFGFILE,
            )
            for entry in config_entries
            if entry.cfg_entry in extras_from_file
        }

        return items_from_file, extras_from_file

    @staticmethod
    def _build_items_from_env(
        config_entries: Sequence[ConfigEntry], opt_from_env: Mapping[str, str]
    ) -> Tuple[Dict[str, ConfigItem], Dict[str, str]]:
        """Extract configuration items from file option and build dict of unconsumed extra items."""
        extras_from_env = {**opt_from_env}
        items_from_env = {
            entry.label: ConfigItem(
                value=entry.convert(extras_from_env.pop(ENVVAR_PREFIX + entry.env_var)),
                source=ConfigSource.ENVVAR,
            )
            for entry in config_entries
            if entry.env_var is not None
            and ENVVAR_PREFIX + entry.env_var in extras_from_env
        }

        return items_from_env, extras_from_env

    @staticmethod
    def _build_parser(
        config_entries: Sequence[ConfigEntry],
        parent_parsers: Optional[Sequence[argparse.ArgumentParser]] = None,
    ) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(
            add_help=False,
            argument_default=argparse.SUPPRESS,
            parents=[] if parent_parsers is None else parent_parsers,
        )
        for config_entry in config_entries:
            if config_entry.cli_flags:
                cli_options = {
                    "help": config_entry.description,
                    "dest": config_entry.label,
                }
                if config_entry.cli_kwargs is not None:
                    cli_options = cli_options | config_entry.cli_kwargs
                flags = (
                    [config_entry.cli_flags]
                    if isinstance(config_entry.cli_flags, str)
                    else config_entry.cli_flags
                )
                parser.add_argument(*flags, **cli_options)
        return parser

    @staticmethod
    def _build_items_from_cli(
        config_entries: Sequence[ConfigEntry], opt_from_cli: Sequence[str]
    ) -> Tuple[Dict[str, ConfigItem], List[str]]:
        """Extract configuration items from file option and build dict of unconsumed extra items."""
        parser = ConfigConsumer._build_parser(config_entries)
        parsed_options, extras_from_cli = parser.parse_known_args(opt_from_cli)

        parsed_dict = vars(parsed_options)

        items_from_cli = {}

        for entry in config_entries:
            if entry.label not in parsed_dict:
                continue
            parsed_value = parsed_dict[entry.label]
            if compatibility_isinstance(parsed_dict[entry.label], entry.type):
                value = parsed_value
            else:
                value = entry.convert(parsed_value)

            items_from_cli[entry.label] = ConfigItem(
                value=value, source=ConfigSource.CLIOPT
            )

        return items_from_cli, extras_from_cli

    @staticmethod
    def _build_items_from_set(
        config_entries: Sequence[ConfigEntry], set_values: Optional[Dict[str, Any]]
    ) -> Dict[str, ConfigItem]:
        """Check that provided items match the entries."""
        set_items = {}

        if set_values is None:
            return set_items

        for config_entry in config_entries:
            if (label := config_entry.label) in set_values:
                set_value = set_values[label]
                if not config_entry.validate(set_value):
                    raise exceptions.ConfigurationValidationError(
                        "Configuration entry '{}' has invalid value '{}'".format(
                            label, set_value
                        )
                    )
                set_items[label] = ConfigItem(set_value, ConfigSource.SETTER)

        return set_items


def userstr_to_bool(user_str: str) -> Optional[bool]:
    """Convert a yes/no, on/off or true/false to bool."""
    if user_str.lower().startswith(("y", "t", "on", "1")):
        return True
    if user_str.lower().startswith(("n", "f", "off", "0")):
        return False
    return None


def userinput_to_pathlist(user_input) -> List[pathlib.Path]:
    """Convert a user input to a list of paths"""
    if user_input is None:
        return []
    if isinstance(user_input, str):
        return userinput_to_pathlist(user_input.split(";"))
    if compatibility_isinstance(user_input, Sequence[str]):
        return [pathlib.Path(path_str) for path_str in user_input]
    if compatibility_isinstance(user_input, Sequence[pathlib.Path]):
        return user_input
    raise ValueError(
        "User input '{}' of type '{}' is not convertible to a list of paths".format(
            user_input, type(user_input)
        )
    )


def userinput_to_strlist(user_input) -> List[str]:
    """Convert a user input to a list of paths"""
    if user_input is None:
        return []
    if isinstance(user_input, str):
        return user_input.replace(" ", ";").split(";")
    if compatibility_isinstance(user_input, List[str]):
        return user_input
    if compatibility_isinstance(user_input, Sequence[str]):
        return [input for input in user_input]
    raise ValueError(
        "User input '{}' of type '{}' is not convertible to a list of strings".format(
            user_input, type(user_input)
        )
    )


class InitialConfigConsumer(ConfigConsumer):
    """
    Class implementing first-pass config consumer.

    On first pass, we only detect if there are CLI arguments which disable
    config file or environment variables.
    """

    ignore_cfg: bool
    ignore_env: bool
    config_file: Optional[pathlib.Path]

    @staticmethod
    def config_entries() -> List[ConfigEntry]:
        return [
            ConfigEntry(
                label="ignore_cfg",
                description="Whether to ignore config file options",
                type=bool,
                default=False,
                env_var="IGNORE_CFG_FILE",
                cli_flags="--ignore-config-file",
                cli_kwargs={"action": "store_const", "const": True},
                convert=userstr_to_bool,
                validate=lambda x: isinstance(x, bool),
            ),
            ConfigEntry(
                label="ignore_env",
                description="Whether to ignore environment variables",
                type=bool,
                default=False,
                cli_flags="--ignore-env",
                cli_kwargs={"action": "store_const", "const": True},
                convert=userstr_to_bool,
                validate=lambda x: isinstance(x, bool),
            ),
            ConfigEntry(
                label="config_file",
                description="Path to GPUBACKENDTOOLS configuration file",
                type=Optional[pathlib.Path],
                default=None,
                cli_flags=["-C", "--config-file"],
                env_var="CONFIG_FILE",
                convert=lambda p: None if p is None else pathlib.Path(p),
                validate=lambda p: True if p is None else p.is_file(),
            ),
        ]

    def __init__(
        self,
        env_vars: Optional[Mapping[str, str]] = None,
        cli_args: Optional[Sequence[str]] = None,
    ):
        super().__init__(config_file=None, env_vars=env_vars, cli_args=cli_args)


def get_package_basepath() -> pathlib.Path:
    import gpubackendtools

    return pathlib.Path(gpubackendtools.__file__).parent


class Configuration(ConfigConsumer):
    """
    Class implementing GPUBACKENDTOOLS complete configuration for the library.
    """

    log_level: int
    log_format: str
    file_registry_path: Optional[pathlib.Path]
    file_storage_path: Optional[pathlib.Path]
    file_download_path: Optional[pathlib.Path]
    file_allow_download: bool
    file_integrity_check: str
    file_extra_paths: List[pathlib.Path]
    file_disabled_tags: Optional[List[str]]
    enabled_backends: Optional[List[str]]

    @staticmethod
    def config_entries() -> List[ConfigEntry]:
        return [
            ConfigEntry(
                label="log_level",
                description="Application log level",
                type=int,
                default=logging.WARN,
                cli_flags=["--log-level"],
                env_var="LOG_LEVEL",
                cfg_entry="log-level",
                convert=Configuration._str_to_logging_level,
            ),
            ConfigEntry(
                label="log_format",
                description="Application log format",
                type=Optional[str],
                default=None,
                cli_flags=["--log-format"],
                env_var="LOG_FORMAT",
                cfg_entry="log-format",
                convert=lambda input: input,
                validate=lambda input: input is None or isinstance(input, str),
            ),
            # ConfigEntry(
            #     label="file_registry_path",
            #     description="File Registry path",
            #     type=Optional[pathlib.Path],
            #     default=None,
            #     cli_flags=["--file-registry"],
            #     env_var="FILE_REGISTRY",
            #     cfg_entry="file-registry",
            #     convert=lambda p: None if p is None else pathlib.Path(p),
            #     validate=lambda p: True if p is None else p.is_file(),
            # ),
            # ConfigEntry(
            #     label="file_storage_path",
            #     description="File Manager storage directory (absolute or relative to current working directory, must exist)",
            #     type=Optional[pathlib.Path],
            #     default=(get_package_basepath() / "data") if is_editable_mode else None,
            #     cli_flags=["--storage-dir"],
            #     env_var="FILE_STORAGE_DIR",
            #     cfg_entry="file-storage-dir",
            #     convert=lambda p: None if p is None else pathlib.Path(p).absolute(),
            #     validate=lambda p: True if p is None else p.is_dir(),
            # ),
            # ConfigEntry(
            #     label="file_download_path",
            #     description="File Manager download directory (absolute, or relative to storage_path)",
            #     type=Optional[pathlib.Path],
            #     default=(get_package_basepath() / "data") if is_editable_mode else None,
            #     cli_flags=["--download-dir"],
            #     env_var="FILE_DOWNLOAD_DIR",
            #     cfg_entry="file-download-dir",
            #     convert=lambda p: None if p is None else pathlib.Path(p),
            #     validate=lambda p: True
            #     if p is None
            #     else (p.is_dir() if p.is_absolute() else True),
            # ),
            # ConfigEntry(
            #     label="file_allow_download",
            #     description="Whether file manager can download missing files from internet",
            #     type=bool,
            #     default=True,
            #     cli_flags="--file-download",
            #     cli_kwargs={"action": argparse.BooleanOptionalAction},
            #     env_var="FILE_ALLOW_DOWNLOAD",
            #     cfg_entry="file-allow-download",
            #     convert=lambda x: userstr_to_bool(x) if isinstance(x, str) else bool(x),
            # ),
            # ConfigEntry(
            #     label="file_integrity_check",
            #     description="When should th file manager perform integrity checks (never, once, always)",
            #     type=str,
            #     default="once",
            #     cli_flags="--file-integrity-check",
            #     cli_kwargs={"choices": ("never", "once", "always")},
            #     env_var="FILE_INTEGRITY_CHECK",
            #     cfg_entry="file-integrity-check",
            #     validate=lambda x: x in ("never", "once", "always"),
            # ),
            # ConfigEntry(
            #     label="file_extra_paths",
            #     description="Supplementary paths in which GPUBACKENDTOOLS will search for files",
            #     type=Optional[List[pathlib.Path]],
            #     default=[get_package_basepath() / "data"],
            #     cli_flags=["--extra-path"],
            #     cli_kwargs={"action": "append"},
            #     env_var="FILE_EXTRA_PATHS",
            #     cfg_entry="file-extra-paths",
            #     convert=userinput_to_pathlist,
            #     overwrite=lambda old, new: old + new
            #     if old is not None
            #     else new,  # concatenate extra path lists
            # ),
            # ConfigEntry(
            #     label="file_disabled_tags",
            #     description="Tags for which access to associated files is disabled",
            #     type=Optional[List[str]],
            #     default=None,
            #     env_var="DISABLED_TAGS",
            #     convert=userinput_to_strlist,
            #     overwrite=lambda old, new: old + new
            #     if old is not None
            #     else new,  # concatenate tag lists
            # ),
            ConfigEntry(
                label="enabled_backends",
                description="List of backends that must be enabled",
                type=Optional[List[str]],
                default=None,
                cli_flags="--enable-backend",
                cli_kwargs={"action": "append"},
                env_var="ENABLED_BACKENDS",
                cfg_entry="enabled-backends",
                convert=lambda x: [v.lower() for v in userinput_to_strlist(x)],
                validate=lambda x: all(v in KNOWN_BACKENDS for v in x)
                if x is not None
                else True,
            ),
        ]

    def __init__(
        self,
        config_file: Union[os.PathLike, Mapping[str, str], None] = None,
        env_vars: Optional[Mapping[str, str]] = None,
        cli_args: Optional[Sequence[str]] = None,
        set_args: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            config_file=config_file,
            env_vars=env_vars,
            cli_args=cli_args,
            set_args=set_args,
        )

        # Post-init task: read -v and -q options
        self._handle_verbosity()

    @staticmethod
    def _str_to_logging_level(input: Union[str, int]) -> int:
        if isinstance(input, int):
            return input
        as_int_level = logging.getLevelName(input.upper())
        if isinstance(as_int_level, int):
            return as_int_level
        raise exceptions.ConfigurationValidationError(
            "'{}' is not a valid log level.".format(input)
        )

    def _apply_verbosity(self, level):
        self._items["log_level"] = ConfigItem(value=level, source=ConfigSource.CLIOPT)

    @staticmethod
    def _build_verbosity_parser() -> argparse.ArgumentParser:
        """Build a parser to handle -v and -Q options"""
        parser = argparse.ArgumentParser(add_help=False)
        exclusive_groups = parser.add_mutually_exclusive_group()
        exclusive_groups.add_argument(
            "-v",
            "--verbose",
            dest="verbose_count",
            action="count",
            default=0,
            help="Increase the verbosity (can be used multiple times)",
        )
        exclusive_groups.add_argument(
            "-Q",
            "--quiet",
            dest="quiet",
            action="store_true",
            help="Disable all logging outputs",
        )
        return parser

    def _handle_verbosity(self):
        parser = self._build_verbosity_parser()
        parsed_options, self._extra_cli = parser.parse_known_args(self._extra_cli)

        from ..globals import get_logger

        if parsed_options.quiet:
            get_logger().debug(
                "Logger level set to CRITICAL since quiet mode is requested."
            )
            self._apply_verbosity(logging.CRITICAL)
        else:
            if (count := parsed_options.verbose_count) == 0:
                return
            old_level = self.log_level
            new_level = old_level - count * 10
            get_logger().debug(
                "Logger level is decreased from {} to {} since verbose flag was set {} times".format(
                    old_level, new_level, count
                )
            )
            self._apply_verbosity(
                new_level if new_level > logging.DEBUG else logging.DEBUG
            )

    def build_cli_parent_parsers(self) -> List[argparse.ArgumentParser]:
        """Build a Parser that can be used as parent parser from CLI-specific parsers"""
        init_parser = ConfigConsumer._build_parser(
            InitialConfigConsumer.config_entries()
        )
        verbosity_parser = self._build_verbosity_parser()
        complete_parser = ConfigConsumer._build_parser(
            Configuration.config_entries(),
            parent_parsers=[init_parser, verbosity_parser],
        )
        return [complete_parser]


def detect_cfg_file() -> Optional[pathlib.Path]:
    """Test common path locations for config and return highest-priority existing one (if any)."""
    import platformdirs

    from .. import __version_tuple__

    LOCATIONS = [
        pathlib.Path.cwd() / "gpubackendtools.ini",
        platformdirs.user_config_path() / "gpubackendtools.ini",
        platformdirs.site_config_path()
        / "gpubackendtools"
        / "v{}.{}".format(__version_tuple__[0], __version_tuple__[1])
        / "gpubackendtools.ini",
    ]
    from ..globals import get_logger

    for location in LOCATIONS:
        if location.is_file():
            get_logger().debug("Configuration file located in '{}'".format(location))
            return location
        get_logger().debug("Configuration file not found in '{}'".format(location))
    return None



"""Definition of global states (logger, config, fast backend, ...)"""

from __future__ import annotations

import logging
import logging.handlers
import os
import typing

from .utils.config import (
    ConfigConsumer,
    ConfigEntry,
    ConfigSource,
    Configuration,
    InitialConfigConsumer,
    detect_cfg_file,
)
from .exceptions import GPUBACKENDTOOLSException

from .gpubackendtools import BackendsManager


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class GPUBACKENDTOOLSGlobalsInitializedTwice(GPUBACKENDTOOLSException):
    """Exception raised if globals are initialized multiple times."""


class GPUBACKENDTOOLSGlobalsReadOnly(GPUBACKENDTOOLSException):
    """Exception raised when trying to modify the global structure."""


class MultiHandlerTarget:
    """Helper class to transition logger from memory-buffer to stream handler during globals init"""

    handlers: typing.List[logging.Handler]
    level: int

    def __init__(self, level: int, *handlers):
        self.handlers = handlers
        self.level = level

    def handle(self, record):
        if record.levelno >= self.level:
            for handler in self.handlers:
                handler.handle(record)


class ConfigurationSetter:
    """Helper class to define configuration options."""

    _args: typing.Dict[str, typing.Any]
    _entries: typing.Dict[str, ConfigEntry]
    _finalizer: typing.Optional[typing.Callable[[typing.Dict[str, typing.Any]], None]]

    def __init__(
        self,
        finalizer: typing.Optional[
            typing.Callable[[typing.Dict[str, typing.Any]], None]
        ] = None,
    ):
        self._args = {}
        self._entries = {entry.label: entry for entry in Configuration.config_entries()}
        self._finalizer = finalizer

    def enable_backends(self, *backends: str) -> ConfigurationSetter:
        """Enable one or multiple backends"""
        return self._convert_and_set("enabled_backends", backends)

    def set_log_level(self, level: typing.Union[str, int]) -> ConfigurationSetter:
        """Set a specific log level"""
        return self._convert_and_set("log_level", level)

    def set_storage_path(self, path: os.PathLike) -> ConfigurationSetter:
        """Modify the storage path"""
        return self._convert_and_set("file_storage_path", path)

    def set_log_format(self, format: str) -> ConfigurationSetter:
        """Change the log format"""
        return self._convert_and_set("log_format", format)

    def set_file_registry_path(self, registry_path: os.PathLike) -> ConfigurationSetter:
        """Set the file registry to use"""
        return self._convert_and_set("file_registry_path", registry_path)

    def set_file_download_path(self, path: os.PathLike) -> ConfigurationSetter:
        """Set the download path"""
        return self._convert_and_set("file_download_path", path)

    def enable_file_download(self) -> ConfigurationSetter:
        """Authorize the file manager to download missing files"""
        return self._convert_and_set("file_allow_download", True)

    def disable_file_download(self) -> ConfigurationSetter:
        """Authorize the file manager to download missing files"""
        return self._convert_and_set("file_allow_download", False)

    def set_file_integrity_check(self, when: str) -> ConfigurationSetter:
        """Define when integrity checks should be performed (never, once, always)"""
        return self._convert_and_set("file_integrity_check", when)

    def disable_file_tags(self, *tags: str) -> ConfigurationSetter:
        """Disable files associated to given tags"""
        return self._convert_and_set("file_disabled_tags", tags)

    def add_file_extra_paths(
        self, *paths: typing.List[os.PathLike]
    ) -> ConfigurationSetter:
        """Add supplementary research paths to file manager"""
        return self._convert_and_set("file_extra_paths", paths)

    def _convert_and_set(self, label: str, value: typing.Any) -> ConfigurationSetter:
        if Globals().is_initialized:
            get_logger().warning(
                "GPUBACKENDTOOLS configurations is already initialized. Option {}={} might not be taken into account".format(
                    label, value
                )
            )
        self._args[label] = self._entries[label].convert(value)
        return self

    def get_args(self) -> dict[str, typing.Any]:
        """Get a dictionnary of set options."""
        return self._args

    def finalize(self):
        """Finalize GPUBACKENDTOOLS initialization with specified parameters."""
        if self._finalizer is not None:
            self._finalizer(self.get_args())


class Globals(metaclass=Singleton):
    _logger: logging.Logger
    _initial_config: InitialConfigConsumer
    _config: Configuration
    # _file_manager: FileManager
    _config_setter: ConfigurationSetter
    _backends_manager: BackendsManager

    _to_initialize: bool

    def __init__(self):
        """Initiliaze the logger"""
        self._preinit_logger()
        self._preinit_setter()
        super().__setattr__("_to_initialize", True)

    def init(
        self,
        cli_args: typing.Optional[typing.Sequence[typing.Any]] = None,
        set_args: typing.Optional[typing.Dict[str, typing.Any]] = None,
    ):
        """Initialize config, file manager and logger with optional CLI arguments."""
        if not super().__getattribute__("_to_initialize"):
            raise GPUBACKENDTOOLSGlobalsInitializedTwice(
                "GPUBACKENDTOOLS globals are already initialized."
            )
        if set_args is None:
            config_setter = self.get_configuration_setter()
            set_args = config_setter.get_args()
        self._init_config(cli_args=cli_args, set_args=set_args)
        self._postconfig_logger()
        # self._init_file_manager()
        self._init_backends_manager()

        super().__setattr__("_to_initialize", False)
        super().__setattr__("_config_setter", None)

        self.logger.debug("GPUBACKENDTOOLS globals initialized.")

    def reset(self):
        """Reset the global structure."""

        # Remove attributes existing only in initialized globals
        if self.is_initialized:
            super().__delattr__("_initial_config")
            super().__delattr__("_config")
            # super().__delattr__("_file_manager")

        # Remove attributes always existing
        super().__delattr__("_logger")
        super().__delattr__("_config_setter")
        super().__delattr__("_to_initialize")

        # Reinitiliaze the structure
        Globals.__init__(self)

    @property
    def is_initialized(self) -> bool:
        """Whether global properties are initialized."""
        return not super().__getattribute__("_to_initialize")

    @property
    def logger(self) -> logging.Logger:
        return super().__getattribute__("_logger")

    @property
    def config(self) -> Configuration:
        if not self.is_initialized:
            self.init()
        return super().__getattribute__("_config")

    # @property
    # def file_manager(self) -> FileManager:
    #     if not self.is_initialized:
    #         self.init()
    #     return super().__getattribute__("_file_manager")

    @property
    def backends_manager(self) -> BackendsManager:
        if super().__getattribute__("_to_initialize"):
            self.init()
        return super().__getattribute__("_backends_manager")

    def get_configuration_setter(self, reset: bool = False) -> ConfigurationSetter:
        """
        Access a configuration setter.

        raises GPUBACKENDTOOLSGlobalsInitializedTwice if globals are already initialized and reset is False.
        """
        if self.is_initialized:
            if reset:
                self.reset()
            else:
                raise GPUBACKENDTOOLSGlobalsInitializedTwice(
                    "GPUBACKENDTOOLS globals are already initialized. Cannot access a setter."
                )
        return super().__getattribute__("_config_setter")

    def __setattr__(self, name, value):
        raise GPUBACKENDTOOLSGlobalsReadOnly(
            "Cannot set attribute on Globals structure."
        )

    def _preinit_logger(self):
        """Pre-initialize logger."""
        logger = logging.getLogger("gpubackendtools")
        logger.setLevel(logging.DEBUG)
        for handler in logger.handlers:
            logger.removeHandler(handler)

        INITIAL_CAPACITY = 1024  # Log up to 1024 messages until globals are initialized
        handler = logging.handlers.MemoryHandler(capacity=INITIAL_CAPACITY)
        handler.set_name("_gpubackendtools_initial_handler")
        logger.addHandler(handler)
        super().__setattr__("_logger", logger)

    def _preinit_setter(self):
        """Initialize the configuration setter."""
        setter = ConfigurationSetter(finalizer=lambda args: self.init(set_args=args))
        super().__setattr__("_config_setter", setter)

    def _init_config(self, cli_args, set_args):
        """Initialize configurations"""
        import os

        ignores_cfg = InitialConfigConsumer(
            cli_args=cli_args
        )  # Read only CLI args (ignores)

        logger = get_logger()
        if ignores_cfg.ignore_env:
            logger.debug(
                "ConfigInitialization: ignoring environment as requested by command-line options"
            )

        file_cfg = InitialConfigConsumer(
            env_vars=None if ignores_cfg.ignore_env else os.environ, cli_args=cli_args
        )  # Read CLI args (and env if not ignored)

        if file_cfg.ignore_cfg:
            logger.debug(
                "ConfigInitialization: ignoring configuration file as requested by {} options".format(
                    "command-line"
                    if file_cfg.get_item("ignore_cfg")[0].source == ConfigSource.CLIOPT
                    else "environment"
                )
            )

        cfg_file = (
            None
            if file_cfg.ignore_cfg
            else file_cfg.config_file
            if file_cfg.config_file is not None
            else detect_cfg_file()
        )

        if cfg_file is not None:
            logger.debug(
                "ConfigInitialization: using configuration file '{}'.".format(cfg_file)
            )

        _, extra_env_vars, extra_cli_args = file_cfg.get_extras()

        config = Configuration(
            config_file=cfg_file,
            env_vars=extra_env_vars,
            cli_args=extra_cli_args,
            set_args=set_args,
        )
        super().__setattr__("_initial_config", file_cfg)
        super().__setattr__("_config", config)

        logger.debug("ConfigInitialization: final configuration entries are")
        self._log_config(file_cfg)
        self._log_config(config)

    def _log_config(self, config: ConfigConsumer, log_level=logging.DEBUG):
        """Print configuration options in logs"""
        logger = get_logger()
        for item, entry in config.get_items():
            logger.log(
                level=log_level,
                msg=f" {entry.label}={item.value} (from: {str(item.source)})",
            )

    def _postconfig_logger(self):
        """Initialize logger after config is initialized."""
        import sys

        logger: logging.Logger = super().__getattribute__("_logger")
        cfg: Configuration = super().__getattribute__("_config")

        stdout_handler = logging.StreamHandler(stream=sys.stdout)
        stdout_handler.addFilter(lambda record: record.levelno <= logging.INFO)

        stderr_handler = logging.StreamHandler(stream=sys.stderr)
        stderr_handler.addFilter(lambda record: record.levelno > logging.INFO)

        if cfg.log_format is not None:
            formatter = logging.Formatter(fmt=cfg.log_format)
            stdout_handler.setFormatter(formatter)
            stderr_handler.setFormatter(formatter)

        for handler in logger.handlers:
            if handler.get_name() == "_gpubackendtools_initial_handler":
                assert isinstance(handler, logging.handlers.MemoryHandler)
                handler.setTarget(
                    MultiHandlerTarget(cfg.log_level, stdout_handler, stderr_handler)
                )
                handler.close()
                break

        logger.handlers.clear()
        logger.setLevel(cfg.log_level)
        logger.addHandler(stdout_handler)
        logger.addHandler(stderr_handler)

    # def _init_file_manager(self):
    #     cfg: Configuration = super().__getattribute__("_config")
    #     file_manager = FileManager(cfg)
    #     super().__setattr__("_file_manager", file_manager)

    def _init_backends_manager(self):
        cfg: Configuration = super().__getattribute__("_config")
        backends_manager = BackendsManager(enabled_backends=cfg.enabled_backends)
        super().__setattr__("_backends_manager", backends_manager)


def get_logger() -> logging.Logger:
    """Get GPUBACKENDTOOLS logger"""
    return Globals().logger


# def get_file_manager() -> FileManager:
#     """Get GPUBACKENDTOOLS File Manager"""
#     return Globals().file_manager


def get_config() -> Configuration:
    """Get GPUBACKENDTOOLS configuration"""
    return Globals().config


def add_backends(available_backends: typing.Dict[str, Backend]):
    """
    Add a backend to the singleton instance.

    """
    Globals().add_backends(available_backends=available_backends)


def get_backend(backend_name: str) -> Backend:
    """
    Get a backend by its name.

    If the backend name is "cuda", return a CUDA backend if any available.
    If the backend name is "gpu", return a GPU backend if any available.
    """
    if "cuda" in backend_name.split("_"):
        _base = backend_name.split("_")[0]
        return get_first_backend([_base + "_cuda12x", _base + "_cuda11x"])
    if "gpu" in backend_name.split("_"):
        _base = backend_name.split("_")[0]
        return get_backend(_base + "_cuda")

    return Globals().backends_manager.get_backend(backend_name=backend_name)


def has_backend(backend_name: str) -> bool:
    """
    Test if a backend is available.

    If the backend name is "cuda", return true if any of "cuda11x" or "cuda12x" is available.
    If the backend name is "gpu", return true if any GPU backend is available.
    """
    if backend_name == "cuda":
        return has_backend("cuda11x") or has_backend("cuda12x")
    if backend_name == "gpu":
        return has_backend("cuda")
    return Globals().backends_manager.has_backend(backend_name=backend_name)


def get_first_backend(backend_names: typing.Sequence[str]) -> Backend:
    """Get the first available backend from a list of backend names"""
    return Globals().backends_manager.get_first_backend(backend_names)


def initialize(*cli_args):
    """Initialize GPUBACKENDTOOLS configuration, logger and file manager with CLI arguments"""
    Globals().init(*cli_args)


def get_config_setter(reset: bool = False) -> ConfigurationSetter:
    """Get a configuration setter."""
    return Globals().get_configuration_setter(reset=reset)


def reset(quiet: bool = False):
    """Reset global states."""
    if not quiet:
        get_logger().warning(
            "GPUBACKENDTOOLS globals are about to be reset. Objects built up until now should be deleted to prevent unexpected side-effects."
        )
    Globals().reset()


# Initialize the globals singleton when first importing this file
Globals()


__all__ = [
    "Globals",
    "ConfigurationSetter",
    "get_logger",
    # "get_file_manager",
    "get_config",
    "get_config_setter",
    "get_backend",
    "get_first_backend",
    "has_backend",
    "initialize",
]



"""Definition of GPUBACKENDTOOLS package common exceptions"""

try:
    from exceptiongroup import ExceptionGroup
except (ImportError, ModuleNotFoundError):
    ExceptionGroup = ExceptionGroup


class GPUBACKENDTOOLSException(Exception):
    """Base class for GPUBACKENDTOOLS package exceptions."""

    pass


class BackendUnavailableException(GPUBACKENDTOOLSException):
    """Exception raised when the backend is not available."""

    pass


class BackendNotInstalled(BackendUnavailableException):
    """Exception raised when the backend has not been installed"""
    pass


class CudaException(GPUBACKENDTOOLSException):
    """Base class for CUDA-related exceptions."""

    pass


class CuPyException(GPUBACKENDTOOLSException):
    """Base class for CuPy-related exceptions."""

    pass


class MissingDependency(GPUBACKENDTOOLSException):
    """Exception raised when a required dependency is missing."""

    pass



"""
File only present in from-source installations to detect editable mode.

It this file is present, the project storage path and download path are set
to $root/src/gpubackendtools/data by default.
This default is still overwritten by `gpubackendtools.ini`, environment variable or other
ways of enforcing configuration options.
"""



"""GPU Backend Tools."""

# ruff: noqa: E402
try:
    from gpubackendtools._version import (  # pylint: disable=E0401,E0611
        __version__,
        __version_tuple__,
    )

except ModuleNotFoundError:
    from importlib.metadata import PackageNotFoundError, version  # pragma: no cover

    try:
        __version__ = version(__name__)
        __version_tuple__ = tuple(__version__.split("."))
    except PackageNotFoundError:  # pragma: no cover
        __version__ = "unknown"
        __version_tuple__ = (0, 0, 0, "unknown")
    finally:
        del version, PackageNotFoundError

_is_editable: bool
try:
    from . import _editable

    _is_editable = True
    del _editable
except (ModuleNotFoundError, ImportError):
    _is_editable = False

from . import cutils, utils
from .globals import get_backend, get_first_backend, has_backend, Globals

from .pointeradjust import wrapper, pointer_adjust

from .parallelbase import ParallelModuleBase

__all__ = [
    "__version__",
    "__version_tuple__",
    "_is_editable",
    # "amplitude",
    # "cutils",
    # "files",
    # "summation",
    # "trajectory",
    # "utils",
    # "waveform",
    "get_logger",
    "get_config",
    "get_config_setter",
    "get_backend",
    "get_first_backend",
    "get_file_manager",
    "has_backend",
    "ParallelModuleBase",
    "wrapper",
    "pointer_adjust",
    "Globals",
]



cornish_fastb = """
@article{Cornish:2007if,
    author = "Cornish, Neil J. and Littenberg, Tyson B.",
    title = "{Tests of Bayesian Model Selection Techniques for Gravitational Wave Astronomy}",
    eprint = "0704.1808",
    archivePrefix = "arXiv",
    primaryClass = "gr-qc",
    doi = "10.1103/PhysRevD.76.083006",
    journal = "Phys. Rev. D",
    volume = "76",
    pages = "083006",
    year = "2007"
}
"""

robson_triple = """
@article{Robson:2018svj,
    author = "Robson, Travis and Cornish, Neil J. and Tamanini, Nicola and Toonen, Silvia",
    title = "{Detecting hierarchical stellar systems with LISA}",
    eprint = "1806.00500",
    archivePrefix = "arXiv",
    primaryClass = "gr-qc",
    doi = "10.1103/PhysRevD.98.064012",
    journal = "Phys. Rev. D",
    volume = "98",
    number = "6",
    pages = "064012",
    year = "2018"
}
"""

zenodo = """
@software{michael_l_katz_2022_6500434,
  author       = {Michael L. Katz},
  title        = {mikekatz04/GPUBACKENDTOOLS: First official public release!},
  month        = apr,
  year         = 2022,
  publisher    = {Zenodo},
  version      = {v1.0.0},
  doi          = {10.5281/zenodo.6500434},
  url          = {https://doi.org/10.5281/zenodo.6500434}
}
"""



from __future__ import annotations
import dataclasses
import enum
import types
import typing
import abc
from .exceptions import GPUBACKENDTOOLSException


class BackendUnavailableException(GPUBACKENDTOOLSException):
    """Exception raised when the backend is not available."""


class BackendNotInstalled(BackendUnavailableException):
    """Exception raised when the backend has not been installed"""


class MissingDependencies(BackendUnavailableException):
    """Exception raised when the backend has missing dependencies"""

    pip_deps: typing.List[str]
    """List of missing dependencies to install in pip-managed environments"""

    conda_deps: typing.List[str]
    """List of missing dependencies to install in conda-managed environments."""

    def __init__(
        self, *args, pip_deps: typing.List[str], conda_deps: typing.List[str], **kwargs
    ):
        self.pip_deps = pip_deps
        self.conda_deps = conda_deps
        super().__init__(*args, **kwargs)

    def __str__(self) -> str:
        message = super().__str__()
        if self.pip_deps:
            message += """
    If you are using gpubackendtools in an environment managed using pip, run:
        $ pip install {}

""".format(", ".join(self.pip_deps))
        if self.conda_deps:
            message += """
    If you are using gpubackendtools in an environment managed using conda, run:
        $ conda install {}

""".format(", ".join(self.conda_deps))
        return message


class MissingDriver(BackendUnavailableException):
    """Exception raised when backend needs driver-like software to be installed."""


class SoftwareException(BackendUnavailableException):
    """Exception raised due to unexpected software error when loading the backend"""


class MissingHardware(BackendUnavailableException):
    """Exception raised when backend needs unavailable hardware."""


@dataclasses.dataclass
class BackendMethods:
    xp: types.ModuleType


class Backend:
    """Abstract definition of a backend"""

    xp: types.ModuleType
    """Reference to package handling the backend ndarrays (numpy or cupy for now)"""

    class Feature(enum.Flag):
        NUMPY = enum.auto()
        """Flag indicating that backend uses numpy.ndarray"""

        CUPY = enum.auto()
        """Flag indicating that backend uses cupy.ndarray"""

        CUDA = enum.auto()
        """Flag indicating that backend uses CUDA devices and offers a set_cuda_device() method."""

        GPU = enum.auto()
        """Flag indicating that backend uses GPU hardware"""

        NONE = 0
        """Special flag representing no activated feature"""

    features: Feature
    """List of Backend features used by a backend"""

    def __init__(self, name: str, methods: BackendMethods, features: Feature):
        # TODO: is this okay?
        # self.name = name
        self.xp = methods.xp
        self.features = features

    @staticmethod
    def _check_module_installed(backend_name: str, module_name: str):
        """Check that the module containing the backend implementation is installed."""
        import importlib

        try:
            importlib.import_module(module_name)
        except ModuleNotFoundError as e:
            raise BackendNotInstalled(
                "The '{}' backend is not installed.".format(backend_name)
            ) from e

    def supports(self, feature: Feature) -> bool:
        """Check whether a backend supports a given feature"""
        return feature in self.features

    @property
    def uses_gpu(self) -> bool:
        """Shortcut to check if a backend supports GPU"""
        return self.supports(feature=Backend.Feature.GPU)

    @property
    def uses_numpy(self) -> bool:
        """Shortcut to check if a backend makes use of NumPy"""
        return self.supports(feature=Backend.Feature.NUMPY)

    @property
    def uses_cupy(self) -> bool:
        """Shortcut to check if a backend makes use of CuPy"""
        return self.supports(feature=Backend.Feature.CUPY)

    @property
    def uses_cuda(self) -> bool:
        """Shortcut to check whether a backend uses CUDA devices"""
        return self.supports(feature=Backend.Feature.CUDA)

    @property
    def name(self) -> str:
        if not hasattr(self, "_name"):
            raise ValueError("Child class must have _name attribute set.")
        return self._name

    @property
    def backend_name(self) -> str:
        if not hasattr(self, "_backend_name"):
            raise ValueError("Child class must have _backend_name attribute set.")
        return self._backend_name
    
    @property
    def backend_base(self) -> str:
        if "_" in self.name:
            return self.name.split("_")[-1]
        else:
            return self.name


class GPUBACKENDTOOLSBackend:
    get_ll: typing.Callable[(...), None]
    fill_global: typing.Callable[(...), None]

    def __init__(self, gpubackendtools_backend_methods):
        # set direct gpubackendtools methods
        # pass rest to general backend
        assert isinstance(
            gpubackendtools_backend_methods, GPUBACKENDTOOLSBackendMethods
        )

        self.get_ll = gpubackendtools_backend_methods.get_ll
        self.fill_global = gpubackendtools_backend_methods.fill_global


class CpuBackend(Backend, abc.ABC):
    """Implementation of the CPU backend"""

    @staticmethod
    @abc.abstractmethod
    def cpu_methods_loader() -> BackendMethods:
        raise not NotImplementedError

    @staticmethod
    def check_numpy() -> typing.ModuleType:
        try:
            import numpy
        except (ModuleNotFoundError, ImportError) as e:
            raise MissingDependencies(
                "'cpu' backend requires numpy", pip_deps=["numpy"], conda_deps=["numpy"]
            ) from e

        return numpy

    def __init__(self):
        """Initialize the CPU backend"""
        if self.backend_name is None:
            raise ValueError("Child class must declare `backend_name` class attribute.")

        self._check_module_installed(self.name, self.backend_name)

        Backend.__init__(
            self,
            name=self.name,
            methods=self.cpu_methods_loader(),
            features=Backend.Feature.NUMPY,
        )


class _CudaBackend(Backend):
    """Implementation of generic CUDA backend"""

    @staticmethod
    def _get_cuda_version() -> typing.Tuple[int, int]:
        """Get the CUDA version or raise an exception"""
        try:
            import pynvml

            pynvml.nvmlInit()
            cuda_version = pynvml.nvmlSystemGetCudaDriverVersion_v2()
        except pynvml.NVMLError_DriverNotLoaded as e:
            raise MissingDriver(
                "CUDA Driver is missing. Ensure you installed it properly."
            ) from e
        except pynvml.NVMLError as e:
            raise SoftwareException(
                "CUDA driver exception: could not detect a CUDA version."
            ) from e
        except AttributeError as e:
            raise SoftwareException(
                "The NVIDIA Management Library (libnvml) does not support the expected method to detect CUDA version."
            ) from e

        cuda_major = cuda_version // 1000
        cuda_minor = (cuda_version % 1000) // 10
        return cuda_major, cuda_minor

    @staticmethod
    def _check_cupy_works(cuda_major: int):
        try:
            import cupy
            import cupy_backends.cuda
        except ImportError as e:
            raise MissingDependencies(
                "CuPy is missing.",
                pip_deps=["cupy-cuda{}x".format(cuda_major)],
                conda_deps=[],
            ) from e

        try:
            cupy_cuda_ver = cupy.cuda.get_local_runtime_version() // 1000
        except RuntimeError as e:
            raise SoftwareException("CuPy could not detect runtime version") from e

        if cupy_cuda_ver != cuda_major:
            raise MissingDependencies(
                "CuPy is installed but supports CUDA {} instead of {}".format(
                    cupy_cuda_ver, cuda_major
                ),
                pip_deps=["cupy-cuda{}x".format(cuda_major)],
                conda_deps=[],
            )

        try:
            _ = cupy.arange(1)
            _ = cupy.cuda.runtime.getDevice()
        except cupy.cuda.compiler.CompileException as e:
            raise MissingDependencies(
                "CuPy fails to run due to missing CUDA Runtime.",
                pip_deps=["nvidia-cuda-runtime-cu{}".format(cuda_major)],
                conda_deps=[],
            ) from e
        except (cupy_backends.cuda.api.runtime.CUDARuntimeError, RuntimeError) as e:
            raise SoftwareException("CuPy could not execute properly.") from e

    @dataclasses.dataclass
    class NvidiaSoLib:
        """Description of a NVidia .so dynamic library"""

        soname: str
        """Dynamic library name (eg libname.so.1)"""

        module_name: str
        """Name of the nvidia module containing the library"""

        pip_pkg: typing.Optional[str] = None
        """Name of a pip-installable package providing that library"""

        conda_pkg: typing.Optional[str] = None
        """Name of a conda-installable package providing that library"""

    @staticmethod
    def _try_import_nvidia_solib(libs: typing.Sequence[NvidiaSoLib]) -> None:
        """Try to load a set of Nvidia dynamic libraries"""
        import ctypes
        import importlib
        import pathlib

        from .exceptions import ExceptionGroup

        try:
            nvidia_root = pathlib.Path(
                importlib.import_module("nvidia").__file__
            ).parent
        except ModuleNotFoundError:
            nvidia_root = None

        failed_idx: typing.List[int] = []
        exceptions = []
        for idx, lib in enumerate(libs):
            try:
                ctypes.cdll.LoadLibrary(lib.soname)
                continue
            except OSError as e:
                exceptions.append(e)

            try:
                if nvidia_root is not None:
                    ctypes.cdll.LoadLibrary(
                        nvidia_root / lib.module_name / "lib" / lib.soname
                    )
                    continue
            except OSError as e:
                exceptions.append(e)

            failed_idx.append(idx)

        if failed_idx:
            raise MissingDependencies(
                "Could not load following NVidia libraries: {}".format(
                    ", ".join([libs[idx].soname for idx in failed_idx])
                ),
                pip_deps=[
                    libs[idx].pip_pkg
                    for idx in failed_idx
                    if libs[idx].pip_pkg is not None
                ],
                conda_deps=[
                    libs[idx].conda_pkg
                    for idx in failed_idx
                    if libs[idx].conda_pkg is not None
                ],
            ) from ExceptionGroup(
                "Following exceptions were raised while trying to load NVidia libraries",
                exceptions,
            )

    @classmethod
    def check_cuda_backend(
        cls,
        name: str,
        backend_module_name: str,
        cuda_min: tuple[int, int],  # Inclusive minimum
        cuda_max: tuple[int, int],  # Exclusive maximum
        module_loader: typing.Callable[[], None],  # Method loading
        dynlib_loader: typing.Optional[typing.Callable[[], None]] = None,
    ) -> BackendMethods:
        """Perform all tests to ensure that a CUDA backend can be used"""

        # 1. Check backend module is installed
        try:
            cls._check_module_installed(name, backend_module_name)
        except BackendNotInstalled as e:
            raise MissingDependencies(
                "GPUBACKENDTOOLS CUDA plugin is missing.",
                pip_deps=["gpubackendtools-cuda{}x".format(cuda_min[0])],
                conda_deps=[],
            ) from e

        # 2. Check CuPy works with expected cuda major version
        cls._check_cupy_works(cuda_major=cuda_min[0])

        # 3. Try to load directly
        try:
            return module_loader()
        except BackendUnavailableException:
            pass

        # 4. Get and check the CUDA version
        cuda_version = cls._get_cuda_version()

        def fmt_version(version: tuple[int, int]) -> str:
            return "{}.{}".format(version[0], version[1])

        if cuda_version < cuda_min:
            raise MissingDriver(
                "Cuda version is below minimum supported version (expected >= {} and < {}, got {})".format(
                    fmt_version(cuda_min),
                    fmt_version(cuda_max),
                    fmt_version(cuda_version),
                )
            )
        if cuda_version >= cuda_max:
            raise MissingDriver(
                "Cuda version is above maximum supported version (expected >= {} and < {}, got {})".format(
                    fmt_version(cuda_min),
                    fmt_version(cuda_max),
                    fmt_version(cuda_version),
                )
            )

        # 5. Retry to load module
        try:
            return module_loader()
        except BackendUnavailableException as e:
            if dynlib_loader is None:
                raise e

        # 6. module_loader failed but dynlib_loader is defined, let's try that
        dynlib_loader()
        return module_loader()

    def set_cuda_device(self, dev: int):
        """Globally sets CUDA device"""
        from cupy.cuda.runtime import setDevice

        setDevice(dev)


class Cuda11xBackend(_CudaBackend, abc.ABC):
    """Implementation of CUDA 11.x backend"""

    @staticmethod
    @abc.abstractmethod
    def cuda11x_module_loader():
        raise NotImplementedError

    @staticmethod
    def cuda11x_dynlib_loader():
        import sys

        if sys.platform == "linux":
            cuda11x_solibs = [
                _CudaBackend.NvidiaSoLib(
                    soname="libcudart.so.11",
                    module_name="cuda_runtime",
                    pip_pkg="nvidia-cuda-runtime-cu11",
                    conda_pkg=None,
                ),
                _CudaBackend.NvidiaSoLib(
                    soname="libcublas.so.11",
                    module_name="cublas",
                    pip_pkg="nvidia-cublas-cu11",
                    conda_pkg=None,
                ),
                _CudaBackend.NvidiaSoLib(
                    soname="libnvJitLink.so.11",
                    module_name="nvjitlink",
                    pip_pkg="nvidia-nvjitlink-cu11",
                    conda_pkg=None,
                ),
                _CudaBackend.NvidiaSoLib(
                    soname="libcusparse.so.11",
                    module_name="cusparse",
                    pip_pkg="nvidia-cusparse-cu11",
                    conda_pkg=None,
                ),
                _CudaBackend.NvidiaSoLib(
                    soname="libnvrtc.so.11",
                    module_name="cuda_nvrtc",
                    pip_pkg="nvidia-cuda-nvrtc-cu11",
                    conda_pkg=None,
                ),
                # TODO: check this??
                # _CudaBackend.NvidiaSoLib(
                #     soname="libcufftw.so.11",
                #     module_name="cufft",
                #     pip_pkg="nvidia-cufft-cu11",
                #     conda_pkg=None,
                # ),
            ]
            _CudaBackend._try_import_nvidia_solib(cuda11x_solibs)

    def __init__(self):
        """Initialize the CPU backend"""

        if self.backend_name is None:
            raise ValueError("Child class must declare `backend_name` class attribute.")

        name = "cuda11x"
        methods = self.check_cuda_backend(
            name=name,
            backend_module_name=self.backend_name,
            cuda_min=(11, 2),
            cuda_max=(12, 0),
            module_loader=Cuda11xBackend.cuda11x_module_loader,
            dynlib_loader=Cuda11xBackend.cuda11x_dynlib_loader,
        )
        Feature = Backend.Feature

        super().__init__(
            name=name,
            methods=methods,
            features=Feature.CUPY | Feature.CUDA | Feature.GPU,
        )


class Cuda12xBackend(_CudaBackend, abc.ABC):
    """Implementation of CUDA 12.x backend"""

    @staticmethod
    @abc.abstractmethod
    def cuda12x_module_loader():
        raise NotImplementedError

    @staticmethod
    def cuda12x_dynlib_loader():
        import sys

        if sys.platform == "linux":
            cuda12x_solibs = [
                _CudaBackend.NvidiaSoLib(
                    soname="libcudart.so.12",
                    module_name="cuda_runtime",
                    pip_pkg="nvidia-cuda-runtime-cu12",
                    conda_pkg=None,
                ),
                _CudaBackend.NvidiaSoLib(
                    soname="libcublas.so.12",
                    module_name="cublas",
                    pip_pkg="nvidia-cublas-cu12",
                    conda_pkg=None,
                ),
                _CudaBackend.NvidiaSoLib(
                    soname="libnvJitLink.so.12",
                    module_name="nvjitlink",
                    pip_pkg="nvidia-nvjitlink-cu12",
                    conda_pkg=None,
                ),
                _CudaBackend.NvidiaSoLib(
                    soname="libcusparse.so.12",
                    module_name="cusparse",
                    pip_pkg="nvidia-cusparse-cu12",
                    conda_pkg=None,
                ),
                _CudaBackend.NvidiaSoLib(
                    soname="libnvrtc.so.12",
                    module_name="cuda_nvrtc",
                    pip_pkg="nvidia-cuda-nvrtc-cu12",
                    conda_pkg=None,
                ),
                # TODO: check this
                # _CudaBackend.NvidiaSoLib(
                #     soname="libcufftw.so.12",
                #     module_name="cufft",
                #     pip_pkg="nvidia-cufft-cu12",
                #     conda_pkg=None,
                # ),
            ]
            _CudaBackend._try_import_nvidia_solib(cuda12x_solibs)

    def __init__(self):
        """Initialize the CPU backend"""

        if self.backend_name is None:
            raise ValueError("Child class must declare `backend_name` class attribute.")

        name = "cuda12x"
        methods = self.check_cuda_backend(
            name=name,
            backend_module_name=self.backend_name,
            cuda_min=(12, 0),
            cuda_max=(13, 0),
            module_loader=self.__class__.cuda12x_module_loader,
            dynlib_loader=self.__class__.cuda12x_dynlib_loader,
        )
        Feature = Backend.Feature

        super().__init__(
            name=name,
            methods=methods,
            features=Feature.CUPY | Feature.CUDA | Feature.GPU,
        )


class BackendStatus:
    """Base class for backend statuses in the backends manager"""


class BackendStatusUnloaded(BackendStatus):
    """Backend is not yet loaded"""


@dataclasses.dataclass
class BackendStatusLoaded(BackendStatus):
    """Backend loaded successfully"""

    instance: Backend


class BackendStatusDisabled(BackendStatus):
    """Backend has been disabled by configuration options"""


@dataclasses.dataclass
class BackendStatusUnavailable:
    """Backend could not be loaded due to exception"""

    reason: BackendUnavailableException


class BackendAccessException(GPUBACKENDTOOLSException):
    """Method raised by BackendManager if requested backend cannot be accessed"""


class BackendsManager:
    """Handles loading and accessing backend instances."""

    _known_backends: typing.Dict = {}
    _registry: typing.Dict[str, BackendStatus]

    @property
    def backend_list(self) -> typing.List[str]:
        """Return the list of backend names"""
        return [name for name in self._registry.keys()]

    def __init__(self, enabled_backends: typing.Optional[typing.Sequence[str]] = None):
        """
        # TODO: how do we handle this as we go to other packages.
        Initialize the backend registry.

        enabled_backends: optional list of backend names which must be loaded, others will be disabled
        """
        self._registry = {}

    def _initialize_backend(
        self, enabled_backends: typing.Optional[typing.Sequence[str]] = None
    ):
        """
        # TODO: how do we handle this as we go to other packages.
        Initialize the backend registry.

        enabled_backends: optional list of backend names which must be loaded, others will be disabled
        """

        self._registry = {
            name: BackendStatusUnloaded() for name in self._known_backends.keys()
        }

        if enabled_backends is not None:
            for backend_name in self.backend_list:
                if backend_name in enabled_backends:
                    status = self._try_loading_backend(backend_name)
                    if isinstance(status, BackendStatusUnavailable):
                        raise BackendAccessException(
                            "Backend '{}' is marked as enabled but cannot be loaded".format(
                                backend_name
                            )
                        ) from status.reason
                else:
                    self._disable_backend(backend_name)

    def _disable_backend(self, backend_name: str):
        """Mark a backend as disabled"""
        self._registry[backend_name] = BackendStatusDisabled()

    def _try_loading_backend(self, backend_name: str) -> BackendStatus:
        """
        Try to load a backend and return its new status.

        If current status is different from UNLOADED, return it directly.
        """
        if backend_name not in self._registry:
            raise ValueError(
                "'{}' is not a valid backend name, expected any of: {}".format(
                    backend_name, ", ".join(self.backend_list)
                )
            )

        if not isinstance(
            current_status := self._registry[backend_name], BackendStatusUnloaded
        ):
            return current_status

        try:
            new_status = BackendStatusLoaded(
                instance=self._known_backends[backend_name]()
            )
        except BackendUnavailableException as e:
            new_status = BackendStatusUnavailable(reason=e)

        self._registry[backend_name] = new_status
        return new_status

    def get_backend(self, backend_name: str) -> Backend:
        """Get a backend instance or raise a BackendAccessException"""
        status = self._try_loading_backend(backend_name=backend_name)
        if isinstance(status, BackendStatusDisabled):
            raise BackendAccessException(
                "Backend '{}' cannot be accessed, it has been disabled.".format(
                    backend_name
                )
            )
        if isinstance(status, BackendStatusUnavailable):
            raise BackendAccessException(
                "Backend '{}' is unavailable. See previous error messages.".format(
                    backend_name
                )
            ) from status.reason

        assert isinstance(status, BackendStatusLoaded)
        return status.instance

    def has_backend(self, backend_name: str) -> bool:
        """Check if a backend is available"""
        return isinstance(
            self._try_loading_backend(backend_name=backend_name), BackendStatusLoaded
        )

    # TODO: need to figure this out!!!
    def get_first_backend(self, backends: typing.Sequence[str]) -> Backend:
        """Get first available backend from a list or raise BackendAccessException if none available"""
        assert len(backends) > 0

        reasons: typing.List[BackendUnavailableException] = []
        for backend_name in backends:
            # check for backend

            run_through = True
            registry_ind = 0
            keys = list(self._registry.keys())
            while run_through and registry_ind < len(keys):
                name = keys[registry_ind]
                if name.split("_")[-1] == backend_name:
                    backend_name = name
                    run_through = False
                registry_ind += 1

            for name in self._registry.keys():
                if name.split("_")[-1] == backend_name:
                    backend_name = name

            if isinstance(
                status := self._try_loading_backend(backend_name=backend_name),
                BackendStatusLoaded,
            ):
                return status.instance
            if isinstance(status, BackendStatusUnavailable):
                reasons.append(status.reason)
            if isinstance(status, BackendStatusDisabled):
                reasons.append(
                    BackendAccessException(
                        "Backend '{}' is disabled.".format(backend_name)
                    )
                )

        from .exceptions import ExceptionGroup

        raise BackendAccessException(
            "Could not access any of the following backends which are either disabled or unavailable: {}".format(
                ", ".join(backends)
            )
        ) from ExceptionGroup(
            "The backends were not available for following reasons.", reasons
        )

    def add_backend(self, name: str, backend: Backend):
        self._known_backends[name] = backend

    def add_backends(self, backends: typing.Dict[str, Backend]):
        for key, value in backends.items():
            self.add_backend(key, value)
        # TODO: configuration added here?
        self._initialize_backend()


@dataclasses.dataclass
class GpuBackendToolConsumer:
    name: str = None
    compiled_backends: typing.Dict[str, Backend] = None


__all__ = ["BackendsManager", "BackendAccessException"]



import warnings


try:
    from cupy.cuda.runtime import setDevice


except (ModuleNotFoundError, ImportError):
    setDevice = None


def cuda_set_device(dev):
    """Globally sets CUDA device

    Args:
        dev (int): CUDA device number.

    """
    if setDevice is not None:
        setDevice(dev)
    else:
        warnings.warn("Setting cuda device, but cupy/cuda not detected.")



# check to see if cupy is available for gpus
import numpy as np

try:
    import cupy as xp

    gpu = True

except (ImportError, ModuleNotFoundError):
    import numpy as np

    gpu = False


def wrapper(*args, **kwargs):
    """Function to convert array and C/C++ class arguments to ptrs

    This function checks the object type. If it is a cupy or numpy array,
    it will determine its pointer by calling the proper attributes. If you design
    a Cython class to be passed through python, it must have a :code:`ptr`
    attribute.

    If you use this function, you must convert input arrays to size_t data type in Cython and
    then properly cast the pointer as it enters the c++ function. See the
    Cython codes
    `here <https://github.com/BlackHolePerturbationToolkit/FastEMRIWaveforms/tree/master/src>`_
    for examples.

    args:
        *args (list): list of the arguments for a function.
        **kwargs (dict): dictionary of keyword arguments to be converted.

    returns:
        Tuple: (targs, tkwargs) where t indicates target (with pointer values
            rather than python objects).

    """
    # declare target containers
    targs = []
    tkwargs = {}

    # args first
    for arg in args:
        if gpu:
            # cupy arrays
            if isinstance(arg, xp.ndarray):
                targs.append(arg.data.mem.ptr)
                continue

        # numpy arrays
        if isinstance(arg, np.ndarray):
            targs.append(arg.__array_interface__["data"][0])
            continue

        try:
            # cython classes
            targs.append(arg.ptr)
            continue
        except AttributeError:
            # regular argument
            targs.append(arg)

    # kwargs next
    for key, arg in kwargs.items():
        if gpu:
            # cupy arrays
            if isinstance(arg, xp.ndarray):
                tkwargs[key] = arg.data.mem.ptr
                continue

        if isinstance(arg, np.ndarray):
            # numpy arrays
            tkwargs[key] = arg.__array_interface__["data"][0]
            continue

        try:
            # cython classes
            tkwargs[key] = arg.ptr
            continue
        except AttributeError:
            # other arguments
            tkwargs[key] = arg

    return (targs, tkwargs)


def pointer_adjust(func):
    """Decorator function for cupy/numpy agnostic cython

    This decorator applies :func:`few.utils.utility.wrapper` to functions
    via the decorator construction.

    If you use this decorator, you must convert input arrays to size_t data type in Cython and
    then properly cast the pointer as it enters the c++ function. See the
    Cython codes
    `here <https://github.com/BlackHolePerturbationToolkit/FastEMRIWaveforms/tree/master/src>`_
    for examples.

    """

    def func_wrapper(*args, **kwargs):
        # get pointers
        targs, tkwargs = wrapper(*args, **kwargs)
        return func(*targs, **tkwargs)

    return func_wrapper
